/*
 * ALSA driver for Panasonic UniPhier series.
 * 
 * Copyright (c) 2013 Panasonic corporation.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/spinlock.h>
#include <linux/vmalloc.h>
#include <linux/kthread.h>
#include <linux/delay.h>
#include <linux/cdev.h>
#include <linux/dma-mapping.h>

#include "mn2ws-pcm.h"

MODULE_AUTHOR("Katsuhiro Suzuki <suzuki.katsuhiro002@jp.panasonic.com>");
MODULE_DESCRIPTION("Panasonic UniPhier PCM Interface Common Driver");
MODULE_LICENSE("GPL");

const char *mn2ws_pcm_pcmif_get_type(struct mn2ws_pcm_pcmif *pcmif)
{
	const char *name = "?";
	
	switch (pcmif->type) {
	case MN2WS_PCMIF_PLAYBACK:
		name = "p";
		break;
	case MN2WS_PCMIF_CAPTURE:
		name = "c";
		break;
	default:
		name = "u";
		break;
	}
	
	return name;
}

int mn2ws_pcm_pcmif_init(struct mn2ws_pcm_dev *d, struct mn2ws_pcm_pcmif *pcmif)
{
	if (!pcmif->desc->enabled) {
		return 0;
	}
	
	spin_lock_init(&pcmif->spin);
	init_waitqueue_head(&pcmif->q);
	
	pcmif->open_trigger = 0;
	pcmif->start_trigger = 0;
	pcmif->stop_trigger = 0;
	pcmif->silent_mode = 1;
	
	//check the necessary callbacks
	if (!pcmif->desc->setup
		|| !pcmif->desc->start
		|| !pcmif->desc->stop
		|| !pcmif->desc->get_hwptr) {
		PRINTF_WARN("ch%d_%s: BUG! one of "
			"setup/start/stop/get_hwptr "
			"is NULL?\n", d->ch, mn2ws_pcm_pcmif_get_type(pcmif));
		snd_BUG();
	}
	
	//initialize the playback DMA buffer
	if (pcmif->desc->phys_addr_buf == 0) {
		//allocate buffer by kernel
		pcmif->buf = dma_alloc_coherent(NULL, 
			max((int)PAGE_SIZE, (int)pcmif->desc->size_buf), 
			&pcmif->paddr_buf, GFP_KERNEL);
	} else {
		//use fixed buffer
		pcmif->paddr_buf = pcmif->desc->phys_addr_buf;
		
		//map the playback DMA buffer in uncached mode
		pcmif->buf = ioremap_nocache(pcmif->paddr_buf, 
			pcmif->desc->size_buf);
	}
	if (pcmif->buf == 0) {
		PRINTF_WARN("ch%d_%s: failed to alloc.\n", 
			d->ch, mn2ws_pcm_pcmif_get_type(pcmif));
		return -EFAULT;
	}
	memset(pcmif->buf, 0, pcmif->desc->size_buf);
	
	PRINTF_NOTE("ch%d_%s: paddr:0x%08x, map:0x%p\n", 
		d->ch, mn2ws_pcm_pcmif_get_type(pcmif), 
		pcmif->paddr_buf, pcmif->buf);
	
	//create transfer thread
	pcmif->thread = kthread_run(pcmif->thread_main, d, 
		"ch%d%s_%s", d->ch, mn2ws_pcm_pcmif_get_type(pcmif), 
		get_pcm_dev_name());
	if (pcmif->thread == NULL) {
		PRINTF_WARN("ch%d_%s: failed to kthread_create.\n", 
			d->ch, mn2ws_pcm_pcmif_get_type(pcmif));
		return -EFAULT;
	}
	
	return 0;
}

int mn2ws_pcm_pcmif_exit(struct mn2ws_pcm_dev *d, struct mn2ws_pcm_pcmif *pcmif)
{
	if (pcmif->thread) {
		kthread_stop(pcmif->thread);
		pcmif->thread = NULL;
	}
	
	if (pcmif->buf) {
		if (pcmif->desc->phys_addr_buf == 0) {
			dma_free_coherent(NULL, 
				max((int)PAGE_SIZE, 
					(int)pcmif->desc->size_buf), 
				pcmif->buf, pcmif->paddr_buf);
		} else {
			iounmap(pcmif->buf);
		}
		pcmif->buf = NULL;
	}
	
	return 0;
}

void mn2ws_pcm_dbg_pcmif_hwbuf(struct mn2ws_pcm_dev *d, struct mn2ws_pcm_pcmif *pcmif)
{
	DPRINTF("ch%d_%s: hw:0x%p-0x%p(%d)\n", 
		d->ch, mn2ws_pcm_pcmif_get_type(pcmif), 
		pcmif->hw.start, pcmif->hw.start + pcmif->hw.len, 
		pcmif->hw.len);
}

void mn2ws_pcm_dbg_pcmif_ringbuf(struct mn2ws_pcm_dev *d, struct mn2ws_pcm_pcmif *pcmif)
{
	DPRINTF("ch%d_%s: hw(r:%05x,w:%05x,rem:%05x) "
		"alsa(r:%05x,w:%05x,rem:%05x)\n", 
		d->ch, mn2ws_pcm_pcmif_get_type(pcmif), 
		(int)get_rp_ringbuf(&pcmif->hw), 
		(int)get_wp_ringbuf(&pcmif->hw), 
		(int)get_remain_ringbuf(&pcmif->hw), 
		(int)get_rp_ringbuf(&pcmif->alsa), 
		(int)get_wp_ringbuf(&pcmif->alsa), 
		(int)get_remain_ringbuf(&pcmif->alsa));
}

void mn2ws_pcm_dbg_pcmif_ringbuf_maxmin(struct mn2ws_pcm_dev *d, struct mn2ws_pcm_pcmif *pcmif)
{
	static unsigned long j = 0;
	static size_t hw_max[8] = {0, }, hw_min[8] = {-1, };
	static size_t alsa_max[8] = {0, }, alsa_min[8] = {-1, };
	
	hw_max[d->ch] = max(hw_max[d->ch], 
		get_remain_ringbuf(&pcmif->hw));
	hw_min[d->ch] = min(hw_min[d->ch], 
		get_remain_ringbuf(&pcmif->hw));
	alsa_max[d->ch] = max(alsa_max[d->ch], 
		get_remain_ringbuf(&pcmif->alsa));
	alsa_min[d->ch] = min(alsa_min[d->ch], 
		get_remain_ringbuf(&pcmif->alsa));
	
	if (time_after(jiffies, j + HZ)) {
		DPRINTF("ch%d_%s: hw(max:%05x,min:%05x) "
			"alsa(max:%05x,min:%05x)\n", 
			d->ch, mn2ws_pcm_pcmif_get_type(pcmif), 
			(int)hw_max[d->ch], 
			(int)hw_min[d->ch], 
			(int)alsa_max[d->ch], 
			(int)alsa_min[d->ch]);
		
		j = jiffies;
		hw_max[d->ch] = 0;
		hw_min[d->ch] = -1;
		alsa_max[d->ch] = 0;
		alsa_min[d->ch] = -1;
	}
}
